@@ -29,6 +29,11 @@ try:
 except ImportError:
     import queue as queue
 
+try:
+    import statsd
+except ImportError:
+    statsd = None
+
 PRECEDENCE_NORMAL = 0
 PRECEDENCE_LOW = 1
 PRECEDENCE_HIGH = 2
@@ -103,17 +108,26 @@ class OptionReqTask(Task):
 
 
 class Connection(object):
-    """A Connection to a Gearman Server."""
+    """A Connection to a Gearman Server.
 
-    log = logging.getLogger("gear.Connection")
+    :arg str client_id: The client ID associated with this connection.
+        It will be appending to the name of the logger (e.g.,
+        gear.Connection.client_id).  Defaults to 'unknown'.
+    """
 
-    def __init__(self, host, port, ssl_key=None, ssl_cert=None, ssl_ca=None):
+    def __init__(self, host, port, ssl_key=None, ssl_cert=None, ssl_ca=None,
+                 client_id='unknown'):
+        self.log = logging.getLogger("gear.Connection.%s" % (client_id,))
         self.host = host
         self.port = port
         self.ssl_key = ssl_key
         self.ssl_cert = ssl_cert
         self.ssl_ca = ssl_ca
 
+        self.use_ssl = False
+        if all([self.ssl_key, self.ssl_cert, self.ssl_ca]):
+            self.use_ssl = True
+
         self.echo_lock = threading.Lock()
         self._init()
 
@@ -158,7 +172,7 @@ class Connection(object):
                 s = None
                 continue
 
-            if all([self.ssl_key, self.ssl_cert, self.ssl_ca]):
+            if self.use_ssl:
                 self.log.debug("Using SSL")
                 s = ssl.wrap_socket(s, ssl_version=ssl.PROTOCOL_TLSv1,
                                     cert_reqs=ssl.CERT_REQUIRED,
@@ -203,17 +217,57 @@ class Connection(object):
         self.disconnect()
         self.connect()
 
+    def sendRaw(self, data):
+        """Send raw data over the socket.
+
+        :arg bytes data The raw data to send
+        """
+        while True:
+            try:
+                self.conn.send(data)
+            except ssl.SSLError as e:
+                if e.errno == ssl.SSL_ERROR_WANT_READ:
+                    continue
+                elif e.errno == ssl.SSL_ERROR_WANT_WRITE:
+                    continue
+                else:
+                    raise
+            break
+
     def sendPacket(self, packet):
         """Send a packet to the server.
 
         :arg Packet packet: The :py:class:`Packet` to send.
         """
         self.log.debug("Sending packet: %s" % packet)
-        self.conn.send(packet.toBinary())
+        self.sendRaw(packet.toBinary())
 
     def _getAdminRequest(self):
         return self.admin_requests.pop(0)
 
+    def _readRawBytes(self, bytes_to_read):
+        while True:
+            try:
+                buff = self.conn.recv(bytes_to_read)
+            except ssl.SSLError as e:
+                if e.errno == ssl.SSL_ERROR_WANT_READ:
+                    continue
+                elif e.errno == ssl.SSL_ERROR_WANT_WRITE:
+                    continue
+                else:
+                    raise
+            break
+
+        bytes_read = len(buff)
+        if self.use_ssl and (bytes_read < bytes_to_read):
+            remaining = self.conn.pending()
+            while remaining and (bytes_read < bytes_to_read):
+                buff += self.conn.recv(bytes_to_read - bytes_read)
+                remaining = self.conn.pending()
+                bytes_read = len(buff)
+
+        return buff
+
     def readPacket(self):
         """Read one packet or administrative response from the server.
 
@@ -229,7 +283,7 @@ class Connection(object):
         admin = None
         admin_request = None
         while True:
-            c = self.conn.recv(1)
+            c = self._readRawBytes(1)
             if not c:
                 return None
             if admin is None:
@@ -258,7 +312,7 @@ class Connection(object):
             is received.
         """
         self.admin_requests.append(request)
-        self.conn.send(request.getCommand())
+        self.sendRaw(request.getCommand())
         complete = request.waitForResponse(timeout)
         if not complete:
             raise TimeoutError()
@@ -339,8 +393,6 @@ class AdminRequest(object):
         The administrative command.
     """
 
-    log = logging.getLogger("gear.AdminRequest")
-
     command = None
     arguments = []
     response = None
@@ -469,15 +521,14 @@ class Packet(object):
 
     :arg bytes code: The Gearman magic code (:py:data:`constants.REQ` or
         :py:data:`constants.RES`)
-    :arg bytes ptype: The packet type (one of the packet types in constasts).
+    :arg bytes ptype: The packet type (one of the packet types in
+        constants).
     :arg bytes data: The data portion of the packet.
     :arg Connection connection: The connection on which the packet
         was received (optional).
     :raises InvalidDataError: If the magic code is unknown.
     """
 
-    log = logging.getLogger("gear.Packet")
-
     def __init__(self, code, ptype, data, connection=None):
         if not isinstance(code, bytes) and not isinstance(code, bytearray):
             raise TypeError("code must be of type bytes or bytearray")
@@ -536,9 +587,10 @@ class Packet(object):
 
 
 class BaseClientServer(object):
-    log = logging.getLogger("gear.BaseClientServer")
-
-    def __init__(self):
+    def __init__(self, client_id='unknown'):
+        self.client_id = convert_to_bytes(client_id)
+        self.log = logging.getLogger("gear.BaseClientServer.%s" %
+                                     (self.client_id,))
         self.running = True
         self.active_connections = []
         self.inactive_connections = []
@@ -627,11 +679,15 @@ class BaseClientServer(object):
         # it and return ASAP and let the connection thread deal with it.
         self.log.debug("Marking %s as disconnected" % conn)
         self.connections_condition.acquire()
-        jobs = conn.related_jobs.values()
-        self.active_connections.remove(conn)
-        self.inactive_connections.append(conn)
-        self.connections_condition.notifyAll()
-        self.connections_condition.release()
+        try:
+            jobs = conn.related_jobs.values()
+            if conn in self.active_connections:
+                self.active_connections.remove(conn)
+            if conn not in self.inactive_connections:
+                self.inactive_connections.append(conn)
+        finally:
+            self.connections_condition.notifyAll()
+            self.connections_condition.release()
         for job in jobs:
             self.handleDisconnect(job)
 
@@ -703,6 +759,7 @@ class BaseClientServer(object):
         """
 
         self.log.debug("Received packet %s" % packet)
+        start = time.time()
         if packet.ptype == constants.JOB_CREATED:
             self.handleJobCreated(packet)
         elif packet.ptype == constants.WORK_COMPLETE:
@@ -775,6 +832,22 @@ class BaseClientServer(object):
             self.handleOptionRes(packet)
         else:
             self.log.error("Received unknown packet: %s" % packet)
+        end = time.time()
+        self.reportTimingStats(packet.ptype, end - start)
+
+    def reportTimingStats(self, ptype, duration):
+        """Report processing times by packet type
+
+        This method is called by handlePacket to report how long
+        processing took for each packet.  The default implementation
+        does nothing.
+
+        :arg bytes ptype: The packet type (one of the packet types in
+            constants).
+        :arg float duration: The time (in seconds) it took to process
+            the packet.
+        """
+        pass
 
     def _defaultPacketHandler(self, packet):
         self.log.error("Received unhandled packet: %s" % packet)
@@ -915,9 +988,11 @@ class BaseClientServer(object):
         # are told to exit.
         self.running = False
         self.connections_condition.acquire()
-        self.connections_condition.notifyAll()
-        os.write(self.wake_write, b'1\n')
-        self.connections_condition.release()
+        try:
+            self.connections_condition.notifyAll()
+            os.write(self.wake_write, b'1\n')
+        finally:
+            self.connections_condition.release()
 
     def _cleanup(self):
         # The second part of the shutdown process where we wait for all
@@ -928,11 +1003,14 @@ class BaseClientServer(object):
             connection.disconnect()
         self.active_connections = []
         self.inactive_connections = []
+        os.close(self.wake_read)
+        os.close(self.wake_write)
 
 
 class BaseClient(BaseClientServer):
-    def __init__(self):
-        super(BaseClient, self).__init__()
+    def __init__(self, client_id='unknown'):
+        super(BaseClient, self).__init__(client_id)
+        self.log = logging.getLogger("gear.BaseClient.%s" % (self.client_id,))
         # A lock to use when sending packets that set the state across
         # all known connections.  Note that it doesn't necessarily need
         # to be used for all broadcasts, only those that affect multi-
@@ -973,7 +1051,8 @@ class BaseClient(BaseClientServer):
             for conn in self.active_connections + self.inactive_connections:
                 if conn.host == host and conn.port == port:
                     raise ConfigurationError("Host/port already specified")
-            conn = Connection(host, port, ssl_key, ssl_cert, ssl_ca)
+            conn = Connection(host, port, ssl_key, ssl_cert, ssl_ca,
+                              self.client_id)
             self.inactive_connections.append(conn)
             self.connections_condition.notifyAll()
         finally:
@@ -1094,12 +1173,16 @@ class Client(BaseClient):
     default event handlers to react to Gearman events.  Be sure to
     call the superclass event handlers so that they may perform
     job-related housekeeping.
+
+    :arg str client_id: The client ID to provide to Gearman.  It will
+        appear in administrative output and be appended to the name of
+        the logger (e.g., gear.Client.client_id).  Defaults to
+        'unknown'.
     """
 
-    log = logging.getLogger("gear.Client")
-
-    def __init__(self):
-        super(Client, self).__init__()
+    def __init__(self, client_id='unknown'):
+        super(Client, self).__init__(client_id)
+        self.log = logging.getLogger("gear.Client.%s" % (self.client_id,))
         self.options = set()
 
     def __repr__(self):
@@ -1392,8 +1475,8 @@ class Client(BaseClient):
         """
 
         job = packet.getJob()
-        job.known = (packet.getArgument(1) == 1)
-        job.running = (packet.getArgument(2) == 1)
+        job.known = (packet.getArgument(1) == '1')
+        job.running = (packet.getArgument(2) == '1')
         job.numerator = packet.getArgument(3)
         job.denominator = packet.getArgument(4)
 
@@ -1457,19 +1540,27 @@ class FunctionRecord(object):
 class Worker(BaseClient):
     """A Gearman worker.
 
-    :arg str worker_id: The worker ID to provide to Gearman (will
-        appear in administrative command output).
+    :arg str client_id: The client ID to provide to Gearman.  It will
+        appear in administrative output and be appended to the name of
+        the logger (e.g., gear.Worker.client_id).
+    :arg str worker_id: The client ID to provide to Gearman.  It will
+        appear in administrative output and be appended to the name of
+        the logger (e.g., gear.Worker.client_id).  This parameter name
+        is deprecated, use client_id instead.
     """
 
-    log = logging.getLogger("gear.Worker")
-
-    def __init__(self, worker_id):
-        self.worker_id = convert_to_bytes(worker_id)
+    def __init__(self, client_id=None, worker_id=None):
+        if not client_id or worker_id:
+            raise Exception("A client_id must be provided")
+        if worker_id:
+            client_id = worker_id
+        super(Worker, self).__init__(client_id)
+        self.log = logging.getLogger("gear.Worker.%s" % (self.client_id,))
+        self.worker_id = client_id
         self.functions = {}
         self.job_lock = threading.Lock()
         self.waiting_for_jobs = 0
         self.job_queue = queue.Queue()
-        super(Worker, self).__init__()
 
     def __repr__(self):
         return '<gear.Worker 0x%x>' % id(self)
@@ -1573,7 +1664,7 @@ class Worker(BaseClient):
         self.broadcast_lock.acquire()
         try:
             # Called immediately after a successful (re-)connection
-            p = Packet(constants.REQ, constants.SET_CLIENT_ID, self.worker_id)
+            p = Packet(constants.REQ, constants.SET_CLIENT_ID, self.client_id)
             conn.sendPacket(p)
             super(Worker, self)._onConnect(conn)
             for f in self.functions.values():
@@ -1582,7 +1673,7 @@ class Worker(BaseClient):
                     p = Packet(constants.REQ, constants.CAN_DO_TIMEOUT, data)
                 else:
                     p = Packet(constants.REQ, constants.CAN_DO, f.name)
-            conn.sendPacket(p)
+                conn.sendPacket(p)
             conn.changeState("IDLE")
         finally:
             self.broadcast_lock.release()
@@ -1650,31 +1741,33 @@ class Worker(BaseClient):
         """
 
         self.job_lock.acquire()
-        while True:
-            connections = self.active_connections[:]
-            now = time.time()
-            ok = True
-            for connection in connections:
-                if connection.state == "GRAB_WAIT":
-                    # Replies to GRAB_JOB should be fast, give up if we've
-                    # been waiting for more than 5 seconds.
-                    if now - connection.state_time > 5:
-                        self._lostConnection(connection)
-                    else:
-                        ok = False
-            if ok:
-                break
-            else:
-                self.job_lock.release()
-                time.sleep(0.1)
-                self.job_lock.acquire()
+        try:
+            while True:
+                connections = self.active_connections[:]
+                now = time.time()
+                ok = True
+                for connection in connections:
+                    if connection.state == "GRAB_WAIT":
+                        # Replies to GRAB_JOB should be fast, give up if we've
+                        # been waiting for more than 5 seconds.
+                        if now - connection.state_time > 5:
+                            self._lostConnection(connection)
+                        else:
+                            ok = False
+                if ok:
+                    break
+                else:
+                    self.job_lock.release()
+                    time.sleep(0.1)
+                    self.job_lock.acquire()
 
-        while self.waiting_for_jobs > 0:
-            self.waiting_for_jobs -= 1
-            self.job_queue.put(None)
+            while self.waiting_for_jobs > 0:
+                self.waiting_for_jobs -= 1
+                self.job_queue.put(None)
 
-        self._updateStateMachines()
-        self.job_lock.release()
+            self._updateStateMachines()
+        finally:
+            self.job_lock.release()
 
     def _shutdown(self):
         super(Worker, self)._shutdown()
@@ -1774,8 +1867,6 @@ class Worker(BaseClient):
 
 
 class BaseJob(object):
-    log = logging.getLogger("gear.Job")
-
     def __init__(self, name, arguments, unique=None, handle=None):
         self.name = convert_to_bytes(name)
         if (not isinstance(arguments, bytes) and
@@ -1843,8 +1934,6 @@ class Job(BaseJob):
         has been submitted to a Gearman server.
     """
 
-    log = logging.getLogger("gear.Job")
-
     def __init__(self, name, arguments, unique=None):
         super(Job, self).__init__(name, arguments, unique)
         self.data = []
@@ -1886,8 +1975,6 @@ class WorkerJob(BaseJob):
         has been submitted to a Gearman server.
     """
 
-    log = logging.getLogger("gear.WorkerJob")
-
     def __init__(self, handle, name, arguments, unique=None):
         super(WorkerJob, self).__init__(name, arguments, unique, handle)
 
@@ -2010,8 +2097,6 @@ class ServerJob(Job):
         has been assigned to a worker.
     """
 
-    log = logging.getLogger("gear.ServerJob")
-
     def __init__(self, handle, name, arguments, client_connection,
                  unique=None):
         super(ServerJob, self).__init__(name, arguments, unique)
@@ -2038,10 +2123,12 @@ class ServerAdminRequest(AdminRequest):
 class ServerConnection(Connection):
     """A Connection to a Gearman Client."""
 
-    def __init__(self, addr, conn):
+    def __init__(self, addr, conn, use_ssl, client_id):
+        self.log = logging.getLogger("gear.ServerConnection.%s" % (client_id,))
         self.host = addr[0]
         self.port = addr[1]
         self.conn = conn
+        self.use_ssl = use_ssl
         self.client_id = None
         self.functions = set()
         self.related_jobs = {}
@@ -2064,9 +2151,21 @@ class Server(BaseClientServer):
     (not for production use).
 
     :arg int port: The TCP port on which to listen.
+    :arg str ssl_key: Path to the SSL private key.
+    :arg str ssl_cert: Path to the SSL certificate.
+    :arg str ssl_ca: Path to the CA certificate.
+    :arg str statsd_host: statsd hostname.  None means disabled
+        (the default).
+    :arg str statsd_port: statsd port (defaults to 8125).
+    :arg str statsd_prefix: statsd key prefix.
+    :arg str client_id: The ID associated with this server.
+        It will be appending to the name of the logger (e.g.,
+        gear.Server.server_id).  Defaults to 'unknown'.
     """
 
-    def __init__(self, port=4730, ssl_key=None, ssl_cert=None, ssl_ca=None):
+    def __init__(self, port=4730, ssl_key=None, ssl_cert=None, ssl_ca=None,
+                 statsd_host=None, statsd_port=8125, statsd_prefix=None,
+                 server_id='unknown'):
         self.port = port
         self.ssl_key = ssl_key
         self.ssl_cert = ssl_cert
@@ -2079,6 +2178,10 @@ class Server(BaseClientServer):
         self.max_handle = 0
         self.connect_wake_read, self.connect_wake_write = os.pipe()
 
+        self.use_ssl = False
+        if all([self.ssl_key, self.ssl_cert, self.ssl_ca]):
+            self.use_ssl = True
+
         for res in socket.getaddrinfo(None, self.port, socket.AF_UNSPEC,
                                       socket.SOCK_STREAM, 0,
                                       socket.AI_PASSIVE):
@@ -2105,7 +2208,19 @@ class Server(BaseClientServer):
         if port == 0:
             self.port = self.socket.getsockname()[1]
 
-        super(Server, self).__init__()
+        super(Server, self).__init__(server_id)
+        self.log = logging.getLogger("gear.Server.%s" % (self.client_id,))
+
+        if statsd_host:
+            if not statsd:
+                self.log.error("Unable to import statsd module")
+                self.statsd = None
+            else:
+                self.statsd = statsd.StatsClient(statsd_host,
+                                                 statsd_port,
+                                                 statsd_prefix)
+        else:
+            self.statsd = None
 
     def _doConnectLoop(self):
         while self.running:
@@ -2113,6 +2228,7 @@ class Server(BaseClientServer):
                 self.connectLoop()
             except Exception:
                 self.log.exception("Exception in connect loop:")
+                time.sleep(1)
 
     def connectLoop(self):
         poll = select.poll()
@@ -2135,19 +2251,22 @@ class Server(BaseClientServer):
                     self.log.debug("Accepting new connection")
                     c, addr = self.socket.accept()
                     self.log.debug("Accepted new connection")
-                    if all([self.ssl_key, self.ssl_cert, self.ssl_ca]):
+                    if self.use_ssl:
                         c = ssl.wrap_socket(c, server_side=True,
                                             keyfile=self.ssl_key,
                                             certfile=self.ssl_cert,
                                             ca_certs=self.ssl_ca,
                                             cert_reqs=ssl.CERT_REQUIRED,
                                             ssl_version=ssl.PROTOCOL_TLSv1)
-                    conn = ServerConnection(addr, c)
+                    conn = ServerConnection(addr, c, self.use_ssl,
+                                            self.client_id)
                     self.connections_condition.acquire()
-                    self.active_connections.append(conn)
-                    self.connections_condition.notifyAll()
-                    os.write(self.wake_write, b'1\n')
-                    self.connections_condition.release()
+                    try:
+                        self.active_connections.append(conn)
+                        self.connections_condition.notifyAll()
+                        os.write(self.wake_write, b'1\n')
+                    finally:
+                        self.connections_condition.release()
 
     def _shutdown(self):
         super(Server, self)._shutdown()
@@ -2156,15 +2275,20 @@ class Server(BaseClientServer):
     def _cleanup(self):
         super(Server, self)._cleanup()
         self.socket.close()
+        os.close(self.connect_wake_read)
+        os.close(self.connect_wake_write)
 
     def _lostConnection(self, conn):
         # Called as soon as a connection is detected as faulty.
         self.log.debug("Marking %s as disconnected" % conn)
         self.connections_condition.acquire()
-        jobs = conn.related_jobs.values()
-        self.active_connections.remove(conn)
-        self.connections_condition.notifyAll()
-        self.connections_condition.release()
+        try:
+            jobs = conn.related_jobs.values()
+            if conn in self.active_connections:
+                self.active_connections.remove(conn)
+        finally:
+            self.connections_condition.notifyAll()
+            self.connections_condition.release()
         for job in jobs:
             if job.worker_connection == conn:
                 # the worker disconnected, alert the client
@@ -2178,6 +2302,7 @@ class Server(BaseClientServer):
             if job.worker_connection:
                 del job.worker_connection.related_jobs[job.handle]
             del self.jobs[job.handle]
+        self._updateStats()
 
     def getQueue(self):
         """Returns a copy of all internal queues in a flattened form.
@@ -2208,27 +2333,36 @@ class Server(BaseClientServer):
                     if handle == job.handle:
                         queue.remove(job)
                         del self.jobs[handle]
-                        request.connection.conn.send(b'OK\n')
+                        self._updateStats()
+                        request.connection.sendRaw(b'OK\n')
                         return
-        request.connection.conn.send(b'ERR UNKNOWN_JOB\n')
+        request.connection.sendRaw(b'ERR UNKNOWN_JOB\n')
 
-    def handleStatus(self, request):
+    def _getFunctionStats(self):
         functions = {}
         for function in self.functions:
             # Total, running, workers
             functions[function] = [0, 0, 0]
         for job in self.jobs.values():
+            if job.name not in functions:
+                functions[job.name] = [0, 0, 0]
             functions[job.name][0] += 1
             if job.running:
                 functions[job.name][1] += 1
         for connection in self.active_connections:
             for function in connection.functions:
+                if function not in functions:
+                    functions[function] = [0, 0, 0]
                 functions[function][2] += 1
+        return functions
+
+    def handleStatus(self, request):
+        functions = self._getFunctionStats()
         for name, values in functions.items():
-            request.connection.conn.send(("%s\t%s\t%s\t%s\n" %
-                                         (name, values[0], values[1],
-                                          values[2])).encode('utf8'))
-        request.connection.conn.send(b'.\n')
+            request.connection.sendRaw(("%s\t%s\t%s\t%s\n" %
+                                       (name, values[0], values[1],
+                                        values[2])).encode('utf8'))
+        request.connection.sendRaw(b'.\n')
 
     def handleWorkers(self, request):
         for connection in self.active_connections:
@@ -2236,10 +2370,10 @@ class Server(BaseClientServer):
             ip = connection.host
             client_id = connection.client_id or '-'
             functions = ' '.join(connection.functions)
-            request.connection.conn.send(("%s %s %s : %s\n" %
-                                          (fd, ip, client_id, functions))
-                                         .encode('utf8'))
-        request.connection.conn.send(b'.\n')
+            request.connection.sendRaw(("%s %s %s : %s\n" %
+                                       (fd, ip, client_id, functions))
+                                       .encode('utf8'))
+        request.connection.sendRaw(b'.\n')
 
     def wakeConnections(self):
         p = Packet(constants.RES, constants.NOOP, b'')
@@ -2248,6 +2382,60 @@ class Server(BaseClientServer):
                 connection.changeState("AWAKE")
                 connection.sendPacket(p)
 
+    def reportTimingStats(self, ptype, duration):
+        """Report processing times by packet type
+
+        This method is called by handlePacket to report how long
+        processing took for each packet.  If statsd is configured,
+        timing and counts are reported with the key
+        "prefix.packet.NAME".
+
+        :arg bytes ptype: The packet type (one of the packet types in
+            constants).
+        :arg float duration: The time (in seconds) it took to process
+            the packet.
+        """
+        if not self.statsd:
+            return
+        ptype = constants.types.get(ptype, 'UNKNOWN')
+        key = 'packet.%s' % ptype
+        self.statsd.timing(key, int(duration * 1000))
+        self.statsd.incr(key)
+
+    def _updateStats(self):
+        if not self.statsd:
+            return
+
+        # prefix.queue.total
+        # prefix.queue.running
+        # prefix.queue.waiting
+        # prefix.workers
+        base_key = 'queue'
+        total = 0
+        running = 0
+        waiting = 0
+        for job in self.jobs.values():
+            total += 1
+            if job.running:
+                running += 1
+            else:
+                waiting += 1
+
+        key = '.'.join([base_key, 'total'])
+        self.statsd.gauge(key, total)
+
+        key = '.'.join([base_key, 'running'])
+        self.statsd.gauge(key, running)
+
+        key = '.'.join([base_key, 'waiting'])
+        self.statsd.gauge(key, waiting)
+
+        workers = 0
+        for connection in self.active_connections:
+            if connection.functions:
+                workers += 1
+        self.statsd.gauge('workers', workers)
+
     def _handleSubmitJob(self, packet, precedence):
         name = packet.getArgument(0)
         unique = packet.getArgument(1)
@@ -2268,6 +2456,7 @@ class Server(BaseClientServer):
             self.normal_queue.append(job)
         elif precedence == PRECEDENCE_LOW:
             self.low_queue.append(job)
+        self._updateStats()
         self.wakeConnections()
 
     def handleSubmitJob(self, packet):
@@ -2288,6 +2477,7 @@ class Server(BaseClientServer):
                         connection.related_jobs[job.handle] = job
                         job.worker_connection = connection
                     job.running = True
+                    self._updateStats()
                     return job
         return None
 
@@ -2350,6 +2540,7 @@ class Server(BaseClientServer):
             del self.jobs[handle]
             del job.client_connection.related_jobs[handle]
             del job.worker_connection.related_jobs[handle]
+            self._updateStats()
 
     def handleSetClientID(self, packet):
         name = packet.getArgument(0)
